import numpy as np
import google.cloud.firestore as firestore
from txp_cloud.common.config import settings
from typing import List, Optional
import logging
from enum import Enum
from datetime import datetime
from datetime import timedelta

configurations_collection_name = settings.firestore.configurations_collection
tenants_collection_name = settings.firestore.tenants_collection
configuration_reference_field_path = "configuration_ref"
configuration_id_field = "configuration_id"
log = logging.getLogger(__name__)
edges_collection_name = settings.firestore.edges_collection
machines_collection_name = settings.firestore.machines_collection


class SignalMode(Enum):
    TIME = 0
    TIME_FFT = 1
    TIME_PSD = 2
    TIME_FFT_PSD = 3
    FFT = 4
    FFT_PSD = 5
    PSD = 6
    IMAGE = 7

    @staticmethod
    def is_image(mode):
        return mode == SignalMode.IMAGE.value

    @staticmethod
    def is_time_only(mode):
        return not SignalMode.is_fft(mode) and not SignalMode.is_psd(mode)

    @staticmethod
    def is_time(mode):
        return mode in [SignalMode.TIME.value, SignalMode.TIME_FFT.value, SignalMode.TIME_FFT_PSD.value,
                        SignalMode.TIME_PSD.value,
                        SignalMode.IMAGE.value]

    @staticmethod
    def is_fft(mode):
        return mode in [SignalMode.TIME_FFT.value, SignalMode.TIME_FFT_PSD.value, SignalMode.FFT.value,
                        SignalMode.FFT_PSD.value]

    @staticmethod
    def is_psd(mode):
        return mode in [SignalMode.TIME_PSD.value, SignalMode.TIME_FFT_PSD.value, SignalMode.PSD.value,
                        SignalMode.FFT_PSD.value]


def get_sampling_frequency(signal, client):
    configuration = pull_configuration_from_firestore(client, signal["configuration_id"], signal["tenant_id"])
    documents = client.collection(edges_collection_name). \
        where(configuration_reference_field_path, "==", configuration.reference). \
        where("logical_id", "==", signal["edge_logical_id"]).get()
    if len(documents) != 1:
        return 1
    edge_document = documents[0]
    return edge_document.get("is_of_kind_ref").get().get("perceptions")[signal["perception_name"]]["sampling_frequency"]


def from_proto_to_json(proto):
    timestamp, index = proto.metadata.package_id.split("_")
    timestamp = int(timestamp)
    index = int(index)
    prev_index = proto.metadata.previous_part_index
    res = []
    for signal in proto.signals:
        row = {
            "data": [],
            "package_timestamp": timestamp,
            "signal_timestamp": signal.timestamp,
            "previous_part_index": prev_index,
            "part_index": index,
            "perception_name": signal.perception_name,
            "edge_logical_id": proto.metadata.edge_descriptor.logical_id,
            "configuration_id": proto.configuration_id,
            "label": "OK",
            "observation_timestamp": proto.metadata.sampling_window.observation_timestamp,
            "gateway_task_id": proto.metadata.sampling_window.gateway_task_id,
            "sampling_window_index": proto.metadata.sampling_window.sampling_window_index,
            "number_of_sampling_windows": proto.metadata.sampling_window.number_of_sampling_windows,
            "tenant_id": proto.metadata.tenant_id
        }
        row["partition_timestamp"] = str(get_partition_utc_date(row["observation_timestamp"]))
        for i, dimension_signal_sample in enumerate(signal.samples):
            row["data"].append({
                "values": list(dimension_signal_sample.samples),
                "index": i
            })
            i += 1
        res.append(row)
    return res


def get_partition_utc_date(observation_timestamp):
    partition_timestamp = datetime.utcfromtimestamp(observation_timestamp // 1e9)
    partition_timestamp = partition_timestamp - timedelta(minutes=partition_timestamp.minute,
                                                          seconds=partition_timestamp.second)
    return partition_timestamp


def merge_signal_chunks(signal_chunks):
    """Merges all signal chunks for a given signal row.

    Args:
        signal_chunks: chunks of a signal, the signal is split in chunks due to MQTT maximum size per package

    Return:
        merged element.
    """

    signal_chunks = sorted(signal_chunks, key=lambda d: d['part_index'])
    data = []
    for i in range(len(signal_chunks[0]["data"])):
        data.append([])
    for i in range(0, len(signal_chunks)):
        for j, dimension_signal_sample in enumerate(signal_chunks[i]["data"]):
            data[j] = np.concatenate((data[j], dimension_signal_sample["values"]), axis=0)
            data[j] = list(data[j])
    signal_chunks[0]["data"] = [{"values": dimension, "index": i} for i, dimension in enumerate(data)]
    signal_chunks[0]["previous_part_index"] = 0
    return signal_chunks[0]


def get_fft_as_np_array(bigquery_fft):
    fft_per_dimensions = []
    for dimension in bigquery_fft:
        fft_dimension = []
        for z in dimension["values"]:
            fft_dimension.append(np.complex128(complex(z["real"], z["imag"])))
        fft_per_dimensions.append(np.array(fft_dimension))

    return np.array(fft_per_dimensions)


def pull_tenant_doc(db: firestore.Client, tenant_id: str) -> Optional[firestore.DocumentSnapshot]:
    """Pull the tenant document from the Firestore database, given
    the tenant_id value"""
    tenant_doc = db.collection(tenants_collection_name).where(
        "tenant_id", "==", tenant_id
    ).get()

    if not tenant_doc:
        log.warning(f"Tenant Document with tenant_id: {tenant_id} not found.")
        return None

    return tenant_doc[0]


def pull_current_configuration_from_firestore(
        db: firestore.Client,
        tenant_id: str
) -> firestore.DocumentSnapshot:
    """Pull down from the Firestore DB the most recent configuration document.

    Args:
        db: The instance of a Firestore authenticated client.
        tenant_id (str): The tenant_id value to download the current configuration.

    Returns:
        The configuration firestore.DocumentReference.
    """
    tenant_doc: firestore.DocumentSnapshot = pull_tenant_doc(
        db, tenant_id
    )

    if not tenant_doc:
        log.warning(f"Configuration Document was not found, because "
                    f"Tenant with tenant_id: {tenant_id} not found.")
        return None

    configuration = db.collection(configurations_collection_name).where(
        "tenant_ref", "==", tenant_doc.reference
    ).order_by(
        "server_timestamp", direction=firestore.Query.DESCENDING
    ).limit(1).get()

    if not configuration:
        log.warning("No Configuration document was found.")
        return None

    return configuration[0]


def pull_configuration_from_firestore(db: firestore.Client, configuration_id, tenant_id) -> firestore.DocumentSnapshot:
    """Pull down from the Firestore DB a configuration document given tenant_id and configuration_id.

    Returns:
        The configuration firestore.DocumentReference.
    """

    tenant_doc: firestore.DocumentSnapshot = pull_tenant_doc(db, tenant_id)

    if not tenant_doc:
        log.warning(f"Configuration Document was not found, because "
                    f"Tenant with tenant_id: {tenant_id} not found.")
        return None

    configuration = db.collection(configurations_collection_name). \
        where("tenant_ref", "==", tenant_doc.reference).where(configuration_id_field, "==", configuration_id). \
        limit(1).get()

    if not configuration:
        log.warning("No Configuration document was found.")
        return None

    return configuration[0]


def pull_docs_from_collection_associated_with_configuration(db: firestore.Client,
                                                            collection_name: str,
                                                            configuration_ref: firestore.DocumentReference) -> \
        List[firestore.DocumentSnapshot]:
    """Returns the documents of a collection that has the configuration value equal to
    the configuration reference received.

    Returns:
        The list of documents found for the query.
    """
    documents = db.collection(collection_name).where(configuration_reference_field_path, "==", configuration_ref).get()

    if not documents:
        log.warning(f"No documents found in collection {collection_name} for configuration {configuration_ref.id}")

    return documents


def get_perception_from_firestore(config_id, tenant_id, edge_logical_id, perception_name, client):
    configuration = pull_configuration_from_firestore(client, config_id, tenant_id)
    documents = client.collection(edges_collection_name). \
        where(configuration_reference_field_path, "==", configuration.reference). \
        where("logical_id", "==", edge_logical_id).get()
    if len(documents) != 1:
        return None
    edge_document = documents[0]
    return edge_document.get("is_of_kind_ref").get().get("perceptions")[perception_name]


def get_all_tenants_from_firestore(db: firestore.Client):
    documents = db.collection(tenants_collection_name).get()
    return [doc.to_dict() for doc in documents]


def get_machines_from_firestore(db: firestore.Client, tenant_id):
    configuration = pull_current_configuration_from_firestore(db, tenant_id)
    documents = db.collection(machines_collection_name). \
        where(configuration_reference_field_path, "==", configuration.reference).get()
    return [doc.to_dict() for doc in documents]


def get_signal_mode_from_firestore(config_id, tenant_id, edge_logical_id, perception_name, client):
    return get_perception_from_firestore(config_id, tenant_id, edge_logical_id, perception_name, client)["mode"]


def get_signal_dimensions_from_firestore(config_id, tenant_id, edge_logical_id, perception_name, client):
    return get_perception_from_firestore(config_id, tenant_id, edge_logical_id, perception_name, client)["dimensions"]
